<figure>
{{ '{{' }}.:{{filename}}.png?nolink{{ '}}' }}
<caption>**cwEPR-Signal von {{ METADATA['sample']['name'] }} (Proben-Nr. [[sa>{{ METADATA['sample']['id'] }}]]) bei {{ '%.1f' %METADATA['parameter']['temperature_control']['temperature']['value'] }} {{ METADATA['parameter']['temperature_control']['temperature']['unit'] }}** {{ '%d' %METADATA['parameter']['signal_channel']['accumulations'] }} Akkumulationen; MW: {{ METADATA['parameter']['bridge']['attenuation']['value'] }} {{ METADATA['parameter']['bridge']['attenuation']['unit'] }}, {{ '%2.4f' %METADATA['parameter']['bridge']['power']['value'] }} {{ METADATA['parameter']['bridge']['power']['unit']}}; {% if METADATA['parameter']['signal_channel']['receiver_gain']['value'] is defined %} Detektorsensitivität: {{ METADATA['parameter']['signal_channel']['receiver_gain']['value']}} {{METADATA['parameter']['signal_channel']['receiver_gain']['unit']}}{% endif %}, Modulationsfrequenz: {{ '%d' %METADATA['parameter']['signal_channel']['modulation_frequency']['value'] }} {{ METADATA['parameter']['signal_channel']['modulation_frequency']['unit'] }}; Modulationsamplitude: {{ METADATA['parameter']['signal_channel']['modulation_amplitude']['value'] }} {{ METADATA['parameter']['signal_channel']['modulation_amplitude']['unit']}}; {% if METADATA['parameter']['signal_channel']['time_constant']['value'] is defined %} Zeitkonstante: {{ METADATA['parameter']['signal_channel']['time_constant']['value']}} {{METADATA['parameter']['signal_channel']['time_constant']['unit']}}, {% endif %}{% if METADATA['parameter']['signal_channel']['conversion_time']['value'] is defined %} Konversionszeit: {{ METADATA['parameter']['signal_channel']['conversion_time']['value'] }} {{ METADATA['parameter']['signal_channel']['conversion_time']['unit'] }}; {% endif %}  Mikrowellenfrequenz: {{ '%2.6f' %METADATA['parameter']['bridge']['mw_frequency']['value'] }} {{ METADATA['parameter']['bridge']['mw_frequency']['unit'] }}. {{ '{{' }}.:{{filename}}.pdf|Bild als PDF-Datei{{ '}}' }}</caption>
</figure>
